from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import ternary
import logging

logging.basicConfig()

logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)

sns.set_style('white')

class FloryHugginsDiagram:
    """
    Use the degree of polymerization of the components and interaction parameters
    between all the components to evaluate the Flory Huggins energy on a grid
    defined by `n`.

    .. note::

     The first component corresponds to the left axis, the second
     to the right axis and the third to the bottom axis.

    :param n: How many points per full line --> for the triangle around n^2/2 points
    :param degree_of_polymerisation_list: contains the degree of polymerisation of the
        three components (or a size equivalent compared to the solvent molecule). The
        list must contain three numbers.

        Example::

         degree_of_polymerisation_list = [700, 500, 600] 

    :param interaction_parameters_dict: Dictionary of parameters, which describe the
        interaction between the different components. The interaction between
        component 1 and 2 is described by the element with kex "chi12" ::

         interaction_parameters_dict = {
             'chi12': 0.004,
             'chi13': 0.004,
             'chi23': 0.004
             }

        Values above zero stand for stronger interactions and therefore favor demixing
        of the components.

        .. note::

         Could be confusing in the beginnings, however component 1 has the python
         index 0, component 2 index 1 and component 3 has index 2. Please use start with 1 to define
         the interaction parameters.


    :param dissociation_parameters_list: This is an optional parameter, which lets you
        introduce additional entropic terms, that depend on the concentration of a
        specific component and its dissociation coefficient. The idea behind is to take
        a for example dissociated dye-counterion pairs into account ::

         dissociation_parameters_list=[
             {
             'index_of_dissociated_component': 2,
             'dissociation_coefficient': 0.1,
             'degree_of_polymerisation': 1
             }
         ]

        It is a list to make it possible to add different dissociated components. Be
        careful, lots of assumptions are made here, first suggested by Hellebust et al.
        in DOI: 10.1021/ma021442t.
    """
    def __init__(self,
                 n=100,
                 degree_of_polymerisation_list=None,
                 interaction_parameters_dict=None,
                 dissociation_parameters_list=None,
                 example_values=False):

        self.n = n
        self.degree_of_polymerisation_list = degree_of_polymerisation_list
        self.interaction_parameter_dict = interaction_parameters_dict
        self.dissociation_parameters_list = dissociation_parameters_list
        self.group_counter = None
        self.groups_dict = None

        if example_values:
            if degree_of_polymerisation_list is None or interaction_parameters_dict is None:
                self.set_example_values()
            else:
                raise ValueError('If example_values True, degree_of_polymerisation and '
                                 'interaction_parameters are set with example values.')

        self.number_of_components = len(self.degree_of_polymerisation_list)



        # generates the point grid
        ternary_points_dict = {'{}_{}'.format(i, j): 
                                   {
                                       'x': i,
                                       'y': j,
                                       'phi1': 1. / n * j, # j == n --> phi1 == 1.0
                                       'phi2': 1. - 1. / n * (i + j), # j == i == 0 --> phi2 == 1.0
                                       'phi3': 1. / n * i, # i == n --> phi3 == 1.0 
                                       'phi1_e': 1. / n * j,
                                       'phi2_e': 1. - 1. / n * (i + j),
                                       'phi3_e': 1. / n * i,
                                   } for i in range(n + 1) for j in range(n + 1 - i)} 

        self.ternary_df = pd.DataFrame.from_dict(ternary_points_dict, orient='index')

        logger.info('Degrees of polymerisation: {}'.format(self.degree_of_polymerisation_list))
        logger.info('Interaction parameters: {}'.format(self.interaction_parameter_dict))

        if self.dissociation_parameters_list is not None:
            """
            For the dissociated dye-counterion pair an additional species is created 
            and its volume fraction is taken away from the dye component. 
            But to make the ternary diagram still both of the volume fractions are summed together.
            Requires a list of dictionaries, e. g.:
            [{'index_of_dissociated_component': 1, 'dissociation_coefficient': 0.1,
             'degree_of_polymerisation': 1}]
            """
            logger.info('Dissociation parameters:')
            for dissociation_dict in self.dissociation_parameters_list:
                logger.info('{}'.format(dissociation_dict))
                component = 'phi{}'.format(dissociation_dict['index_of_dissociated_component'])
                self.number_of_components += 1
                self.degree_of_polymerisation_list.append(dissociation_dict['degree_of_polymerisation'])
                self.ternary_df.loc[:, 'phi{}_e'.format(self.number_of_components)] = dissociation_dict[
                                                        'dissociation_coefficient'] * self.ternary_df[component]
                self.ternary_df[component + '_e'] -= dissociation_dict['dissociation_coefficient'] * self.ternary_df[component]
        self.check_inputs()

    def check_inputs(self):
        """
        Check all the inputs
        """
        if self.n <= 0:
            raise ValueError('n should be greater than 0')
        if self.dissociation_parameters_list is not None \
                and not isinstance(self.dissociation_parameters_list, list):
            raise ValueError('dissociation_parameters_list should be a list of dictionaries and not'
                             '{}.'.format(type(self.dissociation_parameters_list)))
        number_of_additional_degrees_of_polymerisation = 0
        if self.dissociation_parameters_list is not None:
            number_of_additional_degrees_of_polymerisation = len(self.dissociation_parameters_list)
        if len(self.degree_of_polymerisation_list) != 3 + number_of_additional_degrees_of_polymerisation:
            raise ValueError('For a ternary phase diagram, the number of degrees of polymerisation should be 3. '
                             'Received {} as input. \n'.format(self.degree_of_polymerisation_list))

        components_set = range(1, self.number_of_components + 1)
        for key in self.interaction_parameter_dict.keys():
            # assumption: less than 10 components
            if not set([int(key[3]), int(key[4])]).issubset(components_set):
                raise ValueError('Unvalid interaction parameter, please check {}'.format(key))



    def set_example_values(self):
        """
        If example=True, set example values and ignore the other inputs
        """
        self.degree_of_polymerisation_list = [700, 500, 600]
        self.interaction_parameter_dict = {'chi12': 0.004,
                                           'chi13': 0.004,
                                           'chi23': 0.004}

        logger.info('Set example values.')



    def set_free_energy(self):
        """
        Calculate the free energy for all the points in the grid and add
        them to `ternary_df` with the column name 'free_energy'.
        """
        self.ternary_df.loc[:, 'free_energy'] = 0

        # loop because not every interaction parameter might be set
        # with species that are dissociated the number of components could be higher than 3
        for i in range(1, self.number_of_components+1):
            # entropy contribution
            phi_greater_than_zero = self.ternary_df[self.ternary_df['phi{}_e'.format(i)] > 0].loc[:, 'phi{}_e'.format(i)]
            self.ternary_df.loc[phi_greater_than_zero.index, 'free_energy'] += \
                1. / self.degree_of_polymerisation_list[i-1] * phi_greater_than_zero * np.log(phi_greater_than_zero)

            for j in range(i + 1, self.number_of_components+1):
                # enthalpy contribution (interaction between components)
                if 'chi{}{}'.format(i, j) in self.interaction_parameter_dict.keys():
                    self.ternary_df.loc[:, 'free_energy'] += self.interaction_parameter_dict['chi{}{}'.format(i, j)] * \
                                                             self.ternary_df.loc[:, 'phi{}_e'.format(i)] * \
                                                             self.ternary_df.loc[:, 'phi{}_e'.format(j)]

    def set_position(self):
        """
        Set the position column of the `ternary_df` of the points
        either to "bulk", "edge" or "corner".
        """
        self.ternary_df.loc[:, 'position'] = pd.Series('bulk', index=self.ternary_df.index)
        self.ternary_df.loc[(self.ternary_df['phi1'] == 0) | (self.ternary_df['phi2'] == 0) |
                            (self.ternary_df['phi3'] == 0), 'position'] = 'edge'
        self.ternary_df.loc[(self.ternary_df['phi1'] == 1) | (self.ternary_df['phi2'] == 1) |
                            (self.ternary_df['phi3'] == 1), 'position'] = 'corner'

    def set_unstable_points(self):
        """
        Analyze the energy landscape to determine, which points
        are thermodynamically unstable.
        """
        if 'free_energy' not in self.ternary_df.columns:
            self.set_free_energy()
        if 'position' not in self.ternary_df.columns:
            self.set_position()

        self.ternary_df.loc[:, 'stability'] = pd.Series('stable', index=self.ternary_df.index)

        self.set_secant_energy('horizontal', (-1, 0), (1, 0))
        self.set_secant_energy('backslash', (-1, 1), (1, -1))
        self.set_secant_energy('slash', (0, -1), (0, 1))

        self.ternary_df.loc[(self.ternary_df['position'] == 'bulk') &
                       ((self.ternary_df['free_energy'] > self.ternary_df['secant_horizontal']) |
                        (self.ternary_df['free_energy'] > self.ternary_df['secant_slash']) |
                        (self.ternary_df['free_energy'] > self.ternary_df['secant_backslash'])),
                       'stability'] = 'unstable'

    def count_unstable_neighbours(self):
        """
        Count the number of unstable neighbours.
        """
        if 'stability' not in self.ternary_df.columns:
            self.set_unstable_points()
        neighbour_tuples = [(-1, 0), (-1, 1), (0, 1), (1, 0), (1, -1), (0, -1)]

        unstable_list = [[ self.ternary_df.loc['{}_{}'.format(x + neighbour[0], y + neighbour[1]), 'stability'] for
            neighbour in neighbour_tuples].count('unstable') for x, y in self.ternary_df.loc[self.ternary_df['position'] == 'bulk',['x', 'y']].values]

        self.ternary_df.loc[self.ternary_df['position'] == 'bulk', 'unstable_neighbours'] = unstable_list

    def set_spinodal_driving_force(self):
        """
        Calculate the spinodal driving force and add the column to `ternary_df`.

        .. note::

         It is a measure of how unstable a certain mixed composition is and could
         be relevant for fast solvent quenches.

        """
        if 'stability' not in self.ternary_df.columns:
            self.set_unstable_points()
        self.ternary_df.loc[(self.ternary_df['position'] == 'bulk'), 'secant_min'] = np.min(
            self.ternary_df.loc[(self.ternary_df['position'] == 'bulk'),
                                ['secant_horizontal', 'secant_slash', 'secant_backslash']], axis=1)

        self.ternary_df.loc[(self.ternary_df['stability'] == 'unstable'), 'spinodal_driving_force'] = \
            self.ternary_df.loc[
            (self.ternary_df['position'] == 'bulk'), 'secant_min'] - self.ternary_df.loc[
            (self.ternary_df['position'] == 'bulk'), 'free_energy']

    def set_secant_energy(self, name, neighbour0, neighbour1):
        """
        Calculate the mean energy of two neighbouring points at a
        distance defined by the two vectors `neighbour0` and `neighbour1`.

        :param name: Used to name the column in the df
        :param neighbour0: vector of length 2 pointing to neighbour
        :param neighbour1: vector of length 2 pointing to second neighbour
        """
        self.ternary_df.loc[:, 'secant_{}'.format(name)] = pd.Series(0,
                                                        index=self.ternary_df.index)

        neighbour0_index_list = ['{}_{}'.format(*pos) for pos in zip(self.ternary_df.loc[
                                    self.ternary_df['position'] == 'bulk', 'x'].values + neighbour0[0],
                                    self.ternary_df.loc[self.ternary_df[
                                    'position'] == 'bulk', 'y'].values + neighbour0[1])]



        neighbour1_index_list = ['{}_{}'.format(*pos) for pos in zip(self.ternary_df.loc[
                            self.ternary_df['position'] == 'bulk', 'x'].values + neighbour1[0],
                            self.ternary_df.loc[self.ternary_df[
                            'position'] == 'bulk', 'y'].values + neighbour1[1])]

        self.ternary_df.loc[self.ternary_df['position'] == 'bulk',
                            'secant_{}'.format(name)] = 1. / 2 * (self.ternary_df.loc[neighbour0_index_list,
                            'free_energy'].values + self.ternary_df.loc[neighbour1_index_list, 'free_energy'].values)

    def set_unstable_groups(self):
        """
        Find all unstable points belonging two the same unstable region and
        group them.
        """
        if 'stability' not in self.ternary_df.columns:
            self.set_unstable_points()
        unstable_indices = self.ternary_df[self.ternary_df['stability']=='unstable'].index
        self.ternary_df.loc[:, 'group'] = -1
        self.group_counter = 0
        visited = set([])
        neighbour_tuples = [(-1, 0), (-1, 1), (0, 1), (1, 0), (1, -1), (0, -1)]

        for index in unstable_indices:
            if index not in visited:
                # start new group
                connected_indices = set([index])
                x, y = map(int, index.split('_'))
                new_neighbours = set(
                    ['{}_{}'.format(x + neighbour_tuple[0], y + neighbour_tuple[1]) for neighbour_tuple in
                     neighbour_tuples]).intersection(unstable_indices).intersection(unstable_indices)

                all_neighbours = connected_indices.copy()

                while new_neighbours:
                    connected_indices.update(new_neighbours)

                    all_neighbours.update(set.union(set(['{}_{}'.format(
                        list(map(int, index.split('_')))[0] + neighbour_tuple[0],
                        list(map(int, index.split('_')))[1] + neighbour_tuple[1]) for neighbour_tuple in neighbour_tuples for
                                                         index in new_neighbours]).intersection(unstable_indices)))

                    new_neighbours = all_neighbours - connected_indices

                visited.update(connected_indices)

                self.ternary_df.loc[connected_indices, 'group'] = self.group_counter

                self.group_counter += 1


    #######################################
    # plotting functions                  #
    #######################################

    def plot_free_energy(self, fontsize=16,
                         figsize=(9, 8),
                         alpha=0.6,
                         bottom_axis_label='polymer [3]',
                         right_axis_label='solvent [1]',
                         left_axis_label='polymer [2]',
                         title='Free energy',
                         ax=None):
        """
        Plot the free energy landscape

        :param fontsize: axis label fontsize
        :param figsize: figure size
        :param alpha: tranparency of points
        :param bottom_axis_label: bottom label
        :param right_axis_label:  right label
        :param left_axis_label: left label
        :param title: subplot title
        :param ax: previously defined ax on which to plot
        :return: figure, ax
        """
        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(1, 1, 1)
        if 'free_energy' not in self.ternary_df.columns:
            self.set_free_energy()

        sns.despine(left=True, bottom=True)
        figure, tax = ternary.figure(ax=ax, scale=1)

        cm = plt.cm.get_cmap('RdYlBu')

        # Draw Boundary and Gridlines
        tax.boundary(linewidth=1.5)
        tax.gridlines(color="black", multiple=0.05)

        # Set Axis labels and Title
        tax.set_title(title, fontsize=fontsize+4)
        tax.left_axis_label(left_axis_label, fontsize=fontsize)
        tax.right_axis_label(right_axis_label, fontsize=fontsize)
        tax.bottom_axis_label(bottom_axis_label, fontsize=fontsize)

        tax.scatter(self.ternary_df[['phi3', 'phi1', 'phi2']].values, alpha=alpha, c=self.ternary_df['free_energy'], cmap=cm)
        # Set ticks
        tax.ticks(axis='lbr', linewidth=1, multiple=0.1)

        # Remove default Matplotlib Axes
        tax.clear_matplotlib_ticks()
        figure.tight_layout()

        return figure, tax

    def plot_unstable_points(self,
                             driving_force=False,
                             fontsize=16,
                             figsize=(9, 8),
                             alpha=0.6,
                             bottom_axis_label='polymer [3]',
                             right_axis_label='solvent [1]',
                             left_axis_label='polymer [2]',
                             title='Unstable points',
                             ax=None):
        """
        Plot the unstable points (the points within the spinodal line).

        :param driving_force: Color code (from blue (low) to red (high) the
            driving force for spinodal decomposition.
        :param fontsize: axis label fontsize
        :param figsize: figure size
        :param alpha: tranparency of points
        :param bottom_axis_label: bottom label
        :param right_axis_label:  right label
        :param left_axis_label: left label
        :param title: subplot title
        :param ax: previously defined ax on which to plot
        :return: figure, ax
        """
        sc = None

        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(1, 1, 1)

        if 'stability' not in self.ternary_df.columns:
            self.set_unstable_points()
        sns.despine(left=True, bottom=True)
        figure, tax = ternary.figure(ax=ax, scale=1)

        cm = plt.cm.get_cmap('RdYlBu')

        # Draw Boundary and Gridlines
        tax.boundary(linewidth=1.5)
        tax.gridlines(color="black", multiple=0.05)

        # Set Axis labels and Title
        tax.set_title(title, fontsize=fontsize + 4)
        tax.left_axis_label(left_axis_label, fontsize=fontsize)
        tax.right_axis_label(right_axis_label, fontsize=fontsize)
        tax.bottom_axis_label(bottom_axis_label, fontsize=fontsize)

        # only plot
        unstable_points_to_plot = self.ternary_df.loc[self.ternary_df['stability'] == 'unstable',
                                             ['phi3', 'phi1', 'phi2']].values
        if unstable_points_to_plot.any():
            if driving_force is True:
                if 'spinodal_driving_force' not in self.ternary_df.columns:
                    self.set_spinodal_driving_force()
                sc = tax.scatter(unstable_points_to_plot,
                        c=self.ternary_df.loc[self.ternary_df['stability'] == 'unstable', 'spinodal_driving_force'],
                            cmap=cm,  alpha=alpha)
            else:
                sc = tax.scatter(
                    unstable_points_to_plot,
                    color='blue', alpha=alpha)

        # Set ticks
        tax.ticks(axis='lbr', linewidth=1, multiple=0.1)

        # Remove default Matplotlib Axes
        tax.clear_matplotlib_ticks()
        figure.tight_layout()

        return figure, tax, sc


if __name__ == '__main__':
    # tpd = FloryHugginsDiagram(n=100,
    #                           degree_of_polymerisation_list=[1, 3, 5],
    #                           interaction_parameters_dict={
    #                              'chi12': 1.1,
    #                              'chi13': 0.89,
    #                              'chi23': 1.5
    #                            },
    #                           dissociation_parameters_list=[
    #                                {'index_of_dissociated_component': 2,
    #                                 'dissociation_coefficient': 0.1,
    #                                'degree_of_polymerisation': 1}
    #                          ]
    
    #                           )

    tpd = FloryHugginsDiagram(n=100, example_values=True)
    tpd.set_unstable_groups()


